"""Class to manage the clienting of a Gecko in.touch2 enabled device."""

from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Self

from .async_locator import GeckoAsyncLocator
from .async_spa import GeckoAsyncSpa
from .async_taskman import GeckoAsyncTaskMan
from .automation import (
    GeckoAsyncFacade,
    GeckoAutomationBase,
)
from .const import GeckoConstants
from .spa_events import GeckoSpaEvent
from .spa_state import GeckoSpaState

if TYPE_CHECKING:
    from datetime import datetime

    from .async_spa_descriptor import GeckoAsyncSpaDescriptor

_LOGGER = logging.getLogger(__name__)


class GeckoAsyncSpaMan(ABC, GeckoAsyncTaskMan):
    """
    GeckoAsyncSpaMan class.

    Manage the lifetime of a connection to a spa.

    This class is deliberately an abstract because you must provide your own
    implementation to manage the essential events that are required during operation
    """

    class ReconnectButton(GeckoAutomationBase):
        """Perform reconnect of the spa."""

        def __init__(self, spaman: GeckoAsyncSpaMan) -> None:
            """Initialize the reconnect butto n class."""
            super().__init__(
                spaman.unique_id, "Reconnect", spaman.spa_name, "RECONNECT"
            )
            self._spaman = spaman

        async def async_press(self) -> None:
            """Press the button."""
            await self._spaman.async_reset()

    class PingSensor(GeckoAutomationBase):
        """Sensor with the last ping time, or None."""

        def __init__(self, spaman: GeckoAsyncSpaMan) -> None:
            """Initialize the ping sensor."""
            super().__init__(spaman.unique_id, "Last Ping", spaman.spa_name, "PING")
            self._spaman: GeckoAsyncSpaMan = spaman
            assert self._spaman.spa is not None  # noqa: S101
            self._spaman.spa.watch(self._on_spa_change)
            self._last_ping_at: datetime | None = self._spaman.spa.last_ping_at

        @property
        def state(self) -> datetime | None:
            """The state of the sensor."""
            return self._last_ping_at

        @property
        def unit_of_measurement(self) -> None:
            """The unit of measurement for the sensor, or None."""
            return None

        @property
        def device_class(self) -> str:
            """Device class for the ping sensor."""
            return "timestamp"

        def _on_spa_change(self, *_args: Any) -> None:
            assert self._spaman.spa is not None  # noqa: S101
            if self._last_ping_at == self._spaman.spa.last_ping_at:
                return
            self._last_ping_at = self._spaman.spa.last_ping_at
            self._on_change()

        def __repr__(self) -> str:
            """Get string representation."""
            return f"{self.name}: {self.state}"

    class StatusSensor(GeckoAutomationBase):
        """Sensor with the current state."""

        def __init__(self, spaman: GeckoAsyncSpaMan) -> None:
            """Initialize the status sensor class."""
            super().__init__(spaman.unique_id, "Status", spaman.spa_name, "STATUS")
            self._spaman: GeckoAsyncSpaMan = spaman
            self._state = "Unknown"
            self._last_state = GeckoSpaState.IDLE
            self._last_event = GeckoSpaEvent.SPA_MAN_ENTER

        @property
        def state(self) -> str:
            """The state of the sensor."""
            return self._state

        @property
        def unit_of_measurement(self) -> None:
            """The unit of measurement for the sensor, or None."""
            return None

        @property
        def device_class(self) -> str:
            """Get the device class."""
            return "string"

        @property
        def last_event(self) -> GeckoSpaEvent:
            """Return the last event the sensor was notified of."""
            return self._last_event

        @property
        def spa_state(self) -> GeckoSpaState:
            """Returns the state of the spa at the time of the last event."""
            return self._last_state

        def on_event(self, event: GeckoSpaEvent) -> None:
            """Event handler."""
            self._last_event = event
            self._last_state = self._spaman.spa_state
            self._state = GeckoSpaState.to_string(self.spa_state)
            self._on_change()

        def __repr__(self) -> str:
            """Get string representation."""
            return f"{self.name}: {self.state}"

    class RadioConnectionSensor(GeckoAutomationBase):
        """Sensor with the current radio connection data."""

        def __init__(self, spaman: GeckoAsyncSpaMan) -> None:
            """Initialise the radio sensor class."""
            super().__init__(spaman.unique_id, "RF Signal", spaman.spa_name, "RADIO")
            self.signal: int = 0
            self.set_signal(spaman.spa.signal)

        def set_signal(self, signal: int) -> None:
            """Set signal nstrength."""
            self.signal = signal
            self.signal = min(self.signal, 100)

        @property
        def state(self) -> int:
            """The state of the sensor."""
            return self.signal

        @property
        def device_class(self) -> None:
            """Get the device class."""
            return None

        @property
        def unit_of_measurement(self) -> str:
            """The unit of measurement for the sensor, or None."""
            return "%"

        def __repr__(self) -> str:
            """Get the string representation."""
            return f"{self.name}: {self.state}{self.unit_of_measurement}"

    class RadioChannelSensor(GeckoAutomationBase):
        """Sensor with the current radio connection data."""

        def __init__(self, spaman: GeckoAsyncSpaMan) -> None:
            """Initialize the radio channel sensor."""
            super().__init__(spaman.unique_id, "RF Channel", spaman.spa_name, "CHANNEL")
            self.channel: int = 0
            self.set_channel(spaman.spa.channel)

        def set_channel(self, channel: int) -> None:
            """Set the radio channel."""
            self.channel = channel

        @property
        def state(self) -> int:
            """The state of the sensor."""
            return self.channel

        @property
        def device_class(self) -> None:
            """Get the device class."""
            return None

        @property
        def unit_of_measurement(self) -> None:
            """The unit of measurement for the sensor, or None."""
            return None

        def __repr__(self) -> str:
            """Get the string representation."""
            return f"{self.name}: {self.state}"

    def __init__(self, client_uuid: str, **kwargs: str) -> None:
        """
        Initialize a SpaMan class.

        The preferred pattern is to derive a class from SpaMan and then use it in
        an async with

        ```async with MySpaMan(client_uuid, **kwargs) as SpaMan:```

        If you leave **kwargs empty, then you are responsible for co-ordinating calls
        to methods such as async_locate_spas, async_connect_to_spa and async_connect.

        **kwargs can contain the following items

            spa_address:    The IP address of the spa. Useful if the spa is on a sub-net
            spa_identifier: The ID of a spa (as a string)
            spa_name:       The name of the spa. Useful for status feedback when spa
                            cannot be contacted

        If any of the **kwargs are provided then the spam manager will automatically
        run the sequence discover and connect
        """
        GeckoAsyncTaskMan.__init__(self)
        self._client_id = GeckoConstants.FORMAT_CLIENT_IDENTIFIER.format(
            client_uuid
        ).encode(GeckoConstants.MESSAGE_ENCODING)

        # Optional parameters as supplied from config
        self._spa_address: str | None = kwargs.get("spa_address")
        if self._spa_address == "":
            self._spa_address = None
        self._spa_identifier: str | None = kwargs.get("spa_identifier")
        if self._spa_identifier == "":
            self._spa_identifier = None
        self._spa_name: str | None = kwargs.get("spa_name")
        if self._spa_name == "":
            self._spa_name = None

        self._spa_descriptors: list[GeckoAsyncSpaDescriptor] | None = None
        self._has_descriptors = asyncio.Event()
        self._facade: GeckoAsyncFacade | None = None
        self._facade_state_known = asyncio.Event()
        self._spa: GeckoAsyncSpa | None = None
        self._state_change = asyncio.Event()
        self.set_spa_state(GeckoSpaState.IDLE)

        self._status_sensor: GeckoAsyncSpaMan.StatusSensor | None = None
        self._reconnect_button: GeckoAsyncSpaMan.ReconnectButton | None = None
        self._ping_sensor: GeckoAsyncSpaMan.PingSensor | None = None
        self._radio_sensor: GeckoAsyncSpaMan.RadioConnectionSensor | None = None
        self._channel_sensor: GeckoAsyncSpaMan.RadioChannelSensor | None = None

    ########################################################################
    #
    #   Usage helpers
    #

    async def __aenter__(self) -> Self:
        """Support 'with'."""
        await GeckoAsyncTaskMan.__aenter__(self)
        await self._handle_event(GeckoSpaEvent.SPA_MAN_ENTER)
        self.add_task(self._sequence_pump(), "Sequence Pump", "SPAMAN")
        return self

    async def __aexit__(self, *exc_info: object) -> None:
        """Support 'with'."""
        self.cancel_key_tasks("SPAMAN")
        await self._handle_event(GeckoSpaEvent.SPA_MAN_EXIT, exc_info=exc_info)
        await GeckoAsyncTaskMan.__aexit__(self, exc_info)

    ########################################################################
    #
    #   Public methods
    #

    async def async_reset(self) -> None:
        """Reset the spa manager."""
        try:
            self._spa_descriptors = None
            self._has_descriptors.clear()
            if self._facade is not None:
                await self._facade.disconnect()
                self._facade = None
                self._facade_state_known.clear()
            if self._spa is not None:
                await self._spa.disconnect()
                self._spa = None
            self.set_spa_state(GeckoSpaState.IDLE)
        except Exception:
            _LOGGER.exception("Exception during reset")
            raise

    async def async_locate_spas(
        self, spa_address: str | None = None, spa_identifier: str | None = None
    ) -> list[GeckoAsyncSpaDescriptor] | None:
        """
        Locate spas on this network.

        This API will return a list of GeckoAsyncSpaDescriptor that were
        found on the network. If there are none found, then the return will be
        None. Events will be issued as the locating process proceeds
        """
        try:
            await self._handle_event(GeckoSpaEvent.LOCATING_STARTED)
            locator = GeckoAsyncLocator(
                self,
                self._handle_event,
                spa_address=spa_address,
                spa_identifier=spa_identifier,
            )
            await locator.discover()
            self._spa_descriptors = locator.spas
            self._has_descriptors.set()
            del locator

        finally:
            await self._handle_event(
                GeckoSpaEvent.LOCATING_FINISHED,
                spa_descriptors=self._spa_descriptors,
            )

        return self._spa_descriptors

    async def async_connect_to_spa(
        self, spa_descriptor: GeckoAsyncSpaDescriptor | None
    ) -> GeckoAsyncFacade | None:
        """
        Connect to spa.

        This API will connect to the specified spa using the supplied descriptor.
        """
        assert self._facade is None  # noqa: S101
        if spa_descriptor is None:
            await self._handle_event(GeckoSpaEvent.SPA_NOT_FOUND)
            return None

        try:
            self._spa_name = spa_descriptor.name
            if spa_descriptor.identifier is not None:
                self._spa_identifier = spa_descriptor.identifier_as_string
            await self._handle_event(GeckoSpaEvent.CONNECTION_STARTED)
            self._spa = GeckoAsyncSpa(
                self._client_id, spa_descriptor, self, self._handle_event
            )
            await self._spa.connect()
            # Check state now
            if self.spa_state == GeckoSpaState.SPA_READY:
                self._facade = GeckoAsyncFacade(self._spa, self)
                self._facade_state_known.set()

        finally:
            await self._handle_event(
                GeckoSpaEvent.CONNECTION_FINISHED, facade=self._facade
            )

        return self._facade

    async def async_connect(
        self, spa_identifier: str, spa_address: str | None = None
    ) -> GeckoAsyncFacade | None:
        """
        Connect to spa.

        This API will connect to the specified spa by doing a search with the
        supplied information. This is probably the API most commonly used by
        automation systems to avoid storing too much information in configuration
        """
        _LOGGER.debug("async_connect: ID:%s ADDR:%s", spa_identifier, spa_address)

        spa_descriptors = await self.async_locate_spas(spa_address, spa_identifier)

        assert spa_descriptors is not None  # noqa: S101
        if len(spa_descriptors) == 0:
            await self._handle_event(
                GeckoSpaEvent.SPA_NOT_FOUND,
                spa_address=spa_address,
                spa_identifier=spa_identifier,
            )
            return None

        return await self.async_connect_to_spa(spa_descriptors[0])

    async def async_set_spa_info(
        self,
        spa_address: str | None,
        spa_identifier: str | None,
        spa_name: str | None,
    ) -> None:
        """
        Set the spa information.

        Allows the sequence pump can run the locate and
        connect phases.
        """
        _LOGGER.debug(
            "set_spa_info: ADDR:%s ID:%s NAME:%s", spa_address, spa_identifier, spa_name
        )
        self._spa_address = spa_address
        self._spa_identifier = spa_identifier
        self._spa_name = spa_name
        await self.async_reset()

    async def wait_for_descriptors(self) -> None:
        """Wait for descriptors to be available."""
        await self._has_descriptors.wait()

    async def wait_for_facade(self) -> bool:
        """Wait for facade to be available."""
        await self._facade_state_known.wait()
        return self._facade is not None

    ########################################################################
    #
    #   Properties
    #

    @property
    def unique_id(self) -> str:
        """A unique id for the spa manager."""
        assert self._spa_identifier is not None  # noqa: S101
        return f"{self._spa_identifier.replace(':', '')}"

    @property
    def spa_name(self) -> str:
        """The name of the spa being managed."""
        assert self._spa_name is not None  # noqa: S101
        return self._spa_name

    @property
    def spa_descriptors(self) -> list[GeckoAsyncSpaDescriptor] | None:
        """Get a list of the discovered spas, or None."""
        return self._spa_descriptors

    @property
    def facade(self) -> GeckoAsyncFacade | None:
        """Get the connected facade, or None."""
        return self._facade

    @property
    def spa(self) -> GeckoAsyncSpa | None:
        """Get the spa, or None."""
        return self._spa

    @property
    def spa_state(self) -> GeckoSpaState:
        """Get the spa state."""
        return self._spa_state

    def set_spa_state(self, state: GeckoSpaState) -> None:
        """Set the spa state."""
        _LOGGER.debug("Spa state set to %s", state)
        self._spa_state = state
        self._state_change.set()

    @property
    def status_sensor(self) -> GeckoAsyncSpaMan.StatusSensor | None:
        """Get the status sensor."""
        return self._status_sensor

    @property
    def radio_sensor(self) -> GeckoAsyncSpaMan.RadioConnectionSensor | None:
        """Get the radio sensor."""
        return self._radio_sensor

    @property
    def channel_sensor(self) -> GeckoAsyncSpaMan.RadioChannelSensor | None:
        """Get the channel sensor."""
        return self._channel_sensor

    @property
    def reconnect_button(self) -> GeckoAsyncSpaMan.ReconnectButton | None:
        """Get the reconnect button."""
        return self._reconnect_button

    @property
    def ping_sensor(self) -> GeckoAsyncSpaMan.PingSensor | None:
        """Get the ping sensor."""
        return self._ping_sensor

    def __str__(self) -> str:
        """Get the stringized version."""
        return GeckoSpaState.to_string(self.spa_state)

    ########################################################################
    #
    #   Abstract methods
    #
    @abstractmethod
    async def handle_event(self, event: GeckoSpaEvent, **_kwargs: Any) -> None:
        """Handle the event."""

    ########################################################################
    #
    #   Private methods
    #

    async def _handle_event(self, event: GeckoSpaEvent, **kwargs: Any) -> None:  # noqa: PLR0912, PLR0915
        if (
            self._status_sensor is None
            and self._spa_identifier is not None
            and self._spa_name is not None
        ):
            self._status_sensor = GeckoAsyncSpaMan.StatusSensor(self)
            await self._handle_event(GeckoSpaEvent.CLIENT_HAS_STATUS_SENSOR)

        # Do any pre-processing of the event, such as setting the state or
        # updating the status line
        if event == GeckoSpaEvent.LOCATING_STARTED:
            self.set_spa_state(GeckoSpaState.LOCATING_SPAS)

        elif event == GeckoSpaEvent.LOCATING_FINISHED:
            self.set_spa_state(GeckoSpaState.LOCATED_SPAS)

        elif event == GeckoSpaEvent.SPA_NOT_FOUND:
            self._reconnect_button = GeckoAsyncSpaMan.ReconnectButton(self)
            await self._handle_event(GeckoSpaEvent.CLIENT_HAS_RECONNECT_BUTTON)
            self.set_spa_state(GeckoSpaState.ERROR_SPA_NOT_FOUND)

        elif event == GeckoSpaEvent.CONNECTION_STARTED:
            self.set_spa_state(GeckoSpaState.CONNECTING)
            self._reconnect_button = GeckoAsyncSpaMan.ReconnectButton(self)
            await self._handle_event(GeckoSpaEvent.CLIENT_HAS_RECONNECT_BUTTON)

        elif event == GeckoSpaEvent.CONNECTION_GOT_CHANNEL:
            self._ping_sensor = GeckoAsyncSpaMan.PingSensor(self)
            self._radio_sensor = GeckoAsyncSpaMan.RadioConnectionSensor(self)
            self._channel_sensor = GeckoAsyncSpaMan.RadioChannelSensor(self)
            await self._handle_event(GeckoSpaEvent.CLIENT_HAS_PING_SENSOR)

        elif event == GeckoSpaEvent.CONNECTION_SPA_COMPLETE:
            self.set_spa_state(GeckoSpaState.SPA_READY)

        elif event == GeckoSpaEvent.CONNECTION_FINISHED:
            if self._facade is not None:
                self.set_spa_state(GeckoSpaState.CONNECTED)
                await self._handle_event(GeckoSpaEvent.CLIENT_FACADE_IS_READY)

        elif event == GeckoSpaEvent.RUNNING_PING_NO_RESPONSE:
            if self.spa_state == GeckoSpaState.CONNECTED:
                self.set_spa_state(GeckoSpaState.ERROR_PING_MISSED)
                await self._handle_event(GeckoSpaEvent.CLIENT_FACADE_TEARDOWN)

        elif event == GeckoSpaEvent.RUNNING_PING_RECEIVED:
            if self.spa_state in (
                GeckoSpaState.ERROR_PING_MISSED,
                GeckoSpaState.ERROR_RF_FAULT,
                GeckoSpaState.ERROR_NEEDS_ATTENTION,
            ):
                await self.async_reset()

        elif event == GeckoSpaEvent.RUNNING_SPA_NEEDS_RELOAD:
            await self.async_reset()

        elif event == GeckoSpaEvent.ERROR_RF_ERROR:
            if self.spa_state == GeckoSpaState.CONNECTED:
                self.set_spa_state(GeckoSpaState.ERROR_RF_FAULT)
                await self._handle_event(GeckoSpaEvent.CLIENT_FACADE_TEARDOWN)

        elif event == GeckoSpaEvent.RUNNING_SPA_DISCONNECTED:
            if self.spa_state == GeckoSpaState.CONNECTED:
                self.set_spa_state(GeckoSpaState.IDLE)
                await self._handle_event(GeckoSpaEvent.CLIENT_FACADE_TEARDOWN)

        elif event == GeckoSpaEvent.RUNNING_SPA_PACK_REFRESHED:
            self._radio_sensor.set_signal(self._spa.signal)
            self._channel_sensor.set_channel(self._spa.channel)

        elif event in (
            GeckoSpaEvent.CONNECTION_PROTOCOL_RETRY_TIME_EXCEEDED,
            GeckoSpaEvent.ERROR_PROTOCOL_RETRY_TIME_EXCEEDED,
            GeckoSpaEvent.ERROR_TOO_MANY_RF_ERRORS,
        ):
            self.set_spa_state(GeckoSpaState.ERROR_NEEDS_ATTENTION)

        elif event in (
            GeckoSpaEvent.CONNECTION_CANNOT_FIND_SPA_PACK,
            GeckoSpaEvent.CONNECTION_CANNOT_FIND_CONFIG_VERSION,
            GeckoSpaEvent.CONNECTION_CANNOT_FIND_LOG_VERSION,
        ):
            self.set_spa_state(GeckoSpaState.ERROR_NOT_SUPPORTED)
            data = self._spa.get_snapshot_data()
            _LOGGER.info("Device not supported, %s", data)
            await self._spa.disconnect()

        elif event == GeckoSpaEvent.RUNNING_SPA_WATER_CARE_ERROR:
            assert self.facade is not None  # noqa: S101
            assert self._spa is not None  # noqa: S101
            self.facade.water_care.change_watercare_mode(
                await self._spa.async_get_watercare_mode()
            )

        if self._status_sensor is not None:
            self._status_sensor.on_event(event)

        # Call the abstract method to allow derived classes to do useful work
        # such as disconnecting handlers when the spa needs to reconnect
        # after protocol failure
        await self.handle_event(event, **kwargs)

        # Any post-processing goes here

    async def _sequence_pump(self) -> None:
        """
        SpaMan sequence pump loop.

        This class coordinates running the manager from the
        parameterized constructor and machine state.
        """
        _LOGGER.debug("SpaMan sequence pump started")

        try:
            while True:
                await self._state_change.wait()
                self._state_change.clear()

                if (
                    self.spa_state == GeckoSpaState.IDLE
                    and self._spa_descriptors is None
                ):
                    _LOGGER.debug("Sequence pump did locate")
                    await self.async_locate_spas(self._spa_address)

                if (
                    self.spa_state == GeckoSpaState.LOCATED_SPAS
                    and self._spa_identifier is not None
                    and self._facade is None
                ):
                    _LOGGER.debug("Sequence pump did connect")
                    if self._spa_descriptors is not None:
                        await self.async_connect_to_spa(
                            next(
                                (
                                    d
                                    for d in self._spa_descriptors
                                    if d.identifier_as_string == self._spa_identifier
                                ),
                                None,
                            )
                        )
                    else:
                        await self.async_connect(
                            self._spa_identifier, self._spa_address
                        )

        except asyncio.CancelledError:
            _LOGGER.debug("Spaman sequence pump cancelled")
            raise
        except Exception:
            _LOGGER.exception("Sequence pump caught exception")
            raise



"""Gecko Blowers."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .pump import GeckoPump

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade


class GeckoBlower(GeckoPump):
    """Blowers are based on pumps."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the blower."""
        super().__init__(facade, "Blower", "BL")



"""GeckoLogStruct - A class to manage the LogStruct for 'InXM v2'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 2

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 256

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 511

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return [
            "P1",
            "P2",
            "P3",
            "P4",
            "P5",
            "BLOWER",
            "CP",
            "OZONE",
            "L120",
            "FAN",
            "RH_DutyCycle",
            "VSP1_AvantPlanSpeed",
            "VSP1_MinimumSpeed",
            "VSP3_AvantPlanSpeed",
            "VSP3_MinimumSpeed",
            "LI",
        ]

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return []

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return [
            "AmbiantOHLevel1",
            "AmbiantOHLevel2",
            "ContactorCoilFailErr",
            "Fuse1Err",
            "Fuse2Err",
            "Fuse3Err",
            "KinPumpOff",
            "LearnDetectionErr",
            "NoHeaterCurrentErr",
            "OHCondition",
            "RH_CommErr",
            "RH_DutyCycleErr",
            "RH_RegProbeErr",
            "RegOverHeatFlag",
            "RelayStuck",
            "ScanErr",
            "SupplyErr",
            "ThermistanceErr",
            "rHId",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "Hours": GeckoByteStructAccessor(
                self.struct, "LogStructure/Hours", 256, "ALL"
            ),
            "BrMenu": GeckoBoolStructAccessor(
                self.struct, "LogStructure/BrMenu", 279, 6, None
            ),
            "HeaterPhase": GeckoByteStructAccessor(
                self.struct, "LogStructure/HeaterPhase", 300, None
            ),
            "LearningCurrentOUT1A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT1A",
                280,
                None,
            ),
            "LearningCurrentOUT1B": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT1B",
                282,
                None,
            ),
            "LearningCurrentOUT2A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT2A",
                284,
                None,
            ),
            "LearningCurrentOUT2B": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT2B",
                286,
                None,
            ),
            "LearningCurrentOUT3A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT3A",
                288,
                None,
            ),
            "LearningCurrentOUT4A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT4A",
                290,
                None,
            ),
            "LearningCurrentOUT5A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT5A",
                292,
                None,
            ),
            "LearningCurrentOUT6A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT6A",
                294,
                None,
            ),
            "LearningCurrentOUT7A": GeckoWordStructAccessor(
                self.struct,
                "LogStructure/CurrentLearned/LearningCurrentOUT7A",
                296,
                None,
            ),
            "CurrentHeater": GeckoWordStructAccessor(
                self.struct, "LogStructure/CurrentLearned/CurrentHeater", 298, "ALL"
            ),
            "UD_P1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_P1",
                275,
                14,
                ["OFF", "LOW", "HIGH"],
                2,
                4,
                "ALL",
            ),
            "UD_P2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_P2",
                275,
                12,
                ["OFF", "LOW", "HIGH"],
                2,
                4,
                "Gecko",
            ),
            "UD_P3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_P3",
                275,
                11,
                ["OFF", "HIGH"],
                2,
                2,
                "RD",
            ),
            "UD_P4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_P4",
                275,
                10,
                ["OFF", "HIGH"],
                2,
                2,
                "ALL",
            ),
            "UD_P5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_P5",
                275,
                9,
                ["OFF", "HIGH"],
                2,
                2,
                "ALL",
            ),
            "UD_BLOWER": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_BLOWER",
                275,
                8,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UD_AUX1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_AUX1",
                275,
                5,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UD_AUX2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_AUX2",
                275,
                4,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UD_FB_LT": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_FB_LT",
                275,
                2,
                ["OFF", "FIXCOLOR", "SCAN"],
                2,
                4,
                "ALL",
            ),
            "UD_LIGHT120": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_LIGHT120",
                275,
                1,
                ["OFF", "ON"],
                2,
                2,
                "ALL",
            ),
            "UD_QUIET_PAUSE": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UD_QUIET_PAUSE",
                275,
                0,
                [" ", "QUIET"],
                2,
                2,
                "ALL",
            ),
            "VSP1_UDSpeed": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/VSP1_UDSpeed", 304, "ALL"
            ),
            "VSP3_UDSpeed": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/VSP3_UDSpeed", 305, "ALL"
            ),
            "LightIntensity": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/LightIntensity", 308, "ALL"
            ),
            "P1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P1",
                316,
                0,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P2",
                316,
                2,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P3",
                316,
                4,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P4",
                316,
                6,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P5",
                315,
                0,
                ["OFF", "HIGH"],
                None,
                2,
                None,
            ),
            "BLOWER": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/BLOWER",
                315,
                1,
                ["OFF", "HIGH"],
                None,
                2,
                None,
            ),
            "CP": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/CP",
                315,
                2,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "OZONE": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/OZONE",
                315,
                3,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "L120": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/L120",
                315,
                4,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "FAN": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/FAN",
                315,
                7,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "RH_DutyCycle": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/RH_DutyCycle", 274, None
            ),
            "VSP1_AvantPlanSpeed": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/VSP1_AvantPlanSpeed", 302, None
            ),
            "VSP1_MinimumSpeed": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/VSP1_MinimumSpeed", 306, None
            ),
            "VSP3_AvantPlanSpeed": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/VSP3_AvantPlanSpeed", 303, None
            ),
            "VSP3_MinimumSpeed": GeckoByteStructAccessor(
                self.struct, "LogStructure/DeviceStatus/VSP3_MinimumSpeed", 307, None
            ),
            "RemoteFiltAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteFiltAction",
                309,
                None,
                ["USE_INTERNAL", "IDLE", "STOP", "START", "NEW", "FILTER"],
                None,
                6,
                "ALL",
            ),
            "RemoteFiltDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteFiltDur", 310, "ALL"
            ),
            "RemoteOnzenAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteOnzenAction",
                312,
                None,
                ["USE_INTERNAL", "IDLE", "STOP", "START", "NEW", "FILTER"],
                None,
                6,
                "ALL",
            ),
            "RemoteOnzenDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteOnzenDur", 313, "ALL"
            ),
            "FiltPurge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/FiltPurge", 269, 7, None
            ),
            "FiltClean": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/FiltClean", 269, 5, None
            ),
            "FiltCleanSuspended": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/FilterStatus/FiltCleanSuspended",
                269,
                4,
                None,
            ),
            "FiltPurgeSuspendUD": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/FilterStatus/FiltPurgeSuspendUD",
                279,
                2,
                None,
            ),
            "FiltCleanSuspendUD": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/FilterStatus/FiltCleanSuspendUD",
                279,
                3,
                None,
            ),
            "FilterOT": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/FilterOT", 279, 0, None
            ),
            "CPOT": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/CPOT", 279, 1, None
            ),
            "SWM_Active": GeckoByteStructAccessor(
                self.struct, "LogStructure/SWMStatus/SWM_Active", 267, None
            ),
            "SWM_Risk": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SWMStatus/SWM_Risk",
                268,
                None,
                ["NO", "LOW", "MEDIUM", "HIGH", "EXTREME"],
                None,
                8,
                None,
            ),
            "SWMPurge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SWMPurge", 269, 3, None
            ),
            "SWMPurge_Susp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SWMPurge_Susp", 269, 2, None
            ),
            "DisplayedTemp_G": GeckoTempStructAccessor(
                self.struct, "LogStructure/RegulationStatus/DisplayedTemp_G", 265, None
            ),
            "PumpRun": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/PumpRun", 269, 1, None
            ),
            "CheckFlow": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/CheckFlow", 269, 0, None
            ),
            "Heating": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/Heating", 270, 7, None
            ),
            "CoolingDown": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/CoolingDown", 270, 6, None
            ),
            "KineticPurgeActive": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/KineticPurgeActive",
                270,
                5,
                None,
            ),
            "EconomyActive": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/EconomyActive", 271, 0, None
            ),
            "ProgEconomyActive": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/ProgEconomyActive",
                271,
                2,
                None,
            ),
            "ExtProbeActive": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/ExtProbeActive",
                301,
                0,
                None,
            ),
            "ThermistanceErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/ThermistanceErr", 277, 5, None
            ),
            "ContactorCoilFailErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/ContactorCoilFailErr",
                277,
                4,
                None,
            ),
            "NoHeaterCurrentErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/NoHeaterCurrentErr",
                277,
                3,
                None,
            ),
            "LearnDetectionErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/LearnDetectionErr",
                277,
                2,
                None,
            ),
            "Fuse3Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/Fuse3Err", 277, 1, None
            ),
            "Fuse2Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/Fuse2Err", 277, 0, None
            ),
            "Fuse1Err": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/Fuse1Err", 278, 7, None
            ),
            "SupplyErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SupplyErr", 278, 6, None
            ),
            "ScanErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/ScanErr", 278, 5, None
            ),
            "RegOverHeatFlag": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/RegOverHeatFlag", 278, 4, None
            ),
            "AmbiantOHLevel2": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/AmbiantOHLevel2", 278, 3, None
            ),
            "AmbiantOHLevel1": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/AmbiantOHLevel1", 278, 2, None
            ),
            "RelayStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/RelayStuck", 278, 1, None
            ),
            "rHId": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/rHId", 278, 0, None
            ),
            "KinPumpOff": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/KinPumpOff", 279, 4, None
            ),
            "OHCondition": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/OHCondition", 279, 5, None
            ),
            "RH_WaterTemp": GeckoTempStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_WaterTemp", 261, None
            ),
            "RH_TriacTemp": GeckoTempStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_TriacTemp", 263, None
            ),
            "RH_CommErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_CommErr", 257, 7, None
            ),
            "RH_NotEnoughHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_NotEnoughHeat", 257, 3, None
            ),
            "RH_RegSlope": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_RegSlope", 257, 2, None
            ),
            "RH_LowFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_LowFlo", 257, 0, None
            ),
            "RH_DutyCycleErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_DutyCycleErr", 258, 7, None
            ),
            "RH_NoFloRetry": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_NoFloRetry", 258, 6, None
            ),
            "RH_NoFloTemp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_NoFloTemp", 258, 5, None
            ),
            "RH_NoFloXTries": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_NoFloXTries", 258, 4, None
            ),
            "RH_HR_HL": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HR_HL", 258, 3, None
            ),
            "RH_HW_HL": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HW_HL", 258, 2, None
            ),
            "RH_RegProbeErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_RegProbeErr", 258, 1, None
            ),
            "RH_RegOverHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_RegOverHeat", 258, 0, None
            ),
            "RH_HrTriacPr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HrTriacPr", 259, 7, None
            ),
            "RH_HrTriacOH": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HrTriacOH", 259, 6, None
            ),
            "RH_SWTriacOH": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_SWTriacOH", 259, 5, None
            ),
            "RH_HwTriacOH": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HwTriacOH", 259, 4, None
            ),
            "RH_HrKin": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HrKin", 259, 3, None
            ),
            "RH_HrKinNoFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HrKinNoFlo", 259, 2, None
            ),
            "RH_SwKinTemp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_SwKinTemp", 259, 1, None
            ),
            "RH_HWKinetic": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HWKinetic", 259, 0, None
            ),
            "RH_HeaterDisable": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_HeaterDisable", 260, 4, None
            ),
            "RH_FlowDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_FlowDetected", 260, 3, None
            ),
            "RH_FlowCheck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_FlowCheck", 260, 2, None
            ),
            "RH_HeaterSuspended": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/InthermStatus/RH_HeaterSuspended",
                260,
                1,
                None,
            ),
            "RH_FloCheckReady": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RH_FloCheckReady", 260, 0, None
            ),
        }



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InYT v82'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 82

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return [
            "Out1",
            "Out2",
            "Out3",
            "Out4",
            "Out5",
            "OutHtr",
            "HeatPumpCurrent",
            "HeatPumpFuse",
            "Out6",
            "Out7",
            "Out8",
            "Out9",
            "Out10",
            "Out11",
            "Out12",
            "Direct",
            "Direct2",
            "OutLi",
            "LightActivationSafety",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "ConfigNumber": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/ConfigNumber", 0, "ALL"
            ),
            "Out1": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out1",
                12,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out2",
                13,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out3": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out3",
                14,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out4": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out4",
                15,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out5": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out5",
                16,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "OutHtr": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/OutHtr",
                26,
                None,
                [
                    "NA",
                    "P1H",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "HTR",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "HTR2",
                ],
                None,
                None,
                "ALL",
            ),
            "Out1Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out1Cur", 36, "ALL"
            ),
            "Out2Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out2Cur", 37, "ALL"
            ),
            "Out3Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out3Cur", 38, "ALL"
            ),
            "Out4Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out4Cur", 39, "ALL"
            ),
            "Out5Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out5Cur", 40, "ALL"
            ),
            "OutHtRCur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/OutHtRCur", 50, "ALL"
            ),
            "HeatPumpCurrent": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/HeatPumpCurrent",
                161,
                None,
                [
                    "Not_Set",
                    "0A",
                    "1A",
                    "2A",
                    "3A",
                    "4A",
                    "5A",
                    "6A",
                    "7A",
                    "8A",
                    "9A",
                    "10A",
                    "11A",
                    "12A",
                    "13A",
                    "14A",
                    "15A",
                ],
                None,
                None,
                "ALL",
            ),
            "HeatPumpFuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/HeatPumpFuse",
                163,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out6": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out6",
                17,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out7": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out7",
                18,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out8": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out8",
                19,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out9": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out9",
                20,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out10": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out10",
                21,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out11": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out11",
                22,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Out12": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Out12",
                23,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "Protection",
                    "HTR2",
                    "Waterfall",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "Direct": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Direct",
                24,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "CP"],
                None,
                None,
                "ALL",
            ),
            "Direct2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Direct2",
                25,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "CP"],
                None,
                None,
                "ALL",
            ),
            "Out6Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out6Cur", 41, "ALL"
            ),
            "Out7Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out7Cur", 42, "ALL"
            ),
            "Out8Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out8Cur", 43, "ALL"
            ),
            "Out9Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out9Cur", 44, "ALL"
            ),
            "Out10Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out10Cur", 45, "ALL"
            ),
            "Out11Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out11Cur", 46, "ALL"
            ),
            "Out12Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Out12Cur", 47, "ALL"
            ),
            "DirectCur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/DirectCur", 48, "ALL"
            ),
            "Direct2Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/Direct2Cur", 49, "ALL"
            ),
            "WaterfallAsCP": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/WaterfallAsCP",
                119,
                2,
                "ALL",
            ),
            "WaterfallValveOnCP": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/WaterfallValveOnCP",
                119,
                4,
                "ALL",
            ),
            "OutLi": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutLi",
                79,
                None,
                ["", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "LI"],
                None,
                None,
                None,
            ),
            "LightInts": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LVOutputConfig/LightInts", 80, None
            ),
            "LightActivationSafety": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/LightActivationSafety",
                180,
                None,
                ["Disable", "Enable"],
                None,
                None,
                None,
            ),
            "PumpTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/PumpTimeOut", 54, "ALL"
            ),
            "LightTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/LightTimeOut", 55, "ALL"
            ),
            "L120TimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/L120TimeOut", 56, "ALL"
            ),
            "L120Timer": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/TimeOut/L120Timer",
                81,
                0,
                ["Shared", "Own"],
                None,
                2,
                "ALL",
            ),
            "WaterfallTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/WaterfallTimeOut", 118, "ALL"
            ),
            "AuxTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/AuxTimeOut", 120, "ALL"
            ),
            "FiltInterface": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltInterface",
                32,
                None,
                ["PurgeOnly", "FiltCP", "FiltP1", "FiltP1DurOnly"],
                None,
                None,
                "ALL",
            ),
            "CpUsage": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/CpUsage",
                27,
                None,
                ["STANDARD", "ALWAYS_ON"],
                None,
                None,
                "ALL",
            ),
            "OtOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/OtOption",
                57,
                None,
                ["Disabled", "AlwaysEnabled", "WithSPOver95F"],
                None,
                None,
                "ALL",
            ),
            "PurgeSpeed": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/PurgeSpeed",
                81,
                2,
                ["Lo", "Hi"],
                None,
                2,
                "ALL",
            ),
            "AuxAsBubbleGen": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/AuxAsBubbleGen",
                119,
                1,
                "ALL",
            ),
            "OTTriggerG": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/AdvanceFilterConfig/OTTriggerG", 58, "ALL"
            ),
            "CpOnTimeDuringOT": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/CpOnTimeDuringOT",
                59,
                "ALL",
            ),
            "CpOffTimeDuringOT": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/CpOffTimeDuringOT",
                60,
                "ALL",
            ),
            "FiltOnTimeDuringOT": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOnTimeDuringOT",
                61,
                "ALL",
            ),
            "FiltSuspendTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltSuspendTime",
                62,
                "ALL",
            ),
            "DrainMode": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/DrainMode",
                78,
                None,
                ["NA", "P1", "CP"],
                None,
                None,
                "ALL",
            ),
            "O3Usage": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Usage",
                28,
                None,
                ["Filter", "Always"],
                None,
                None,
                "ALL",
            ),
            "O3Pump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Pump",
                29,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "O3Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Type",
                30,
                None,
                ["Standard", "Toggle"],
                None,
                None,
                "ALL",
            ),
            "O3SuspendTime": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3SuspendTime", 63, "ALL"
            ),
            "SetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/SetpointG",
                1,
                "ALL",
            ),
            "TempUnits": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/TempUnits",
                33,
                None,
                ["F", "C"],
                None,
                None,
                "ALL",
            ),
            "HeaterPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeaterPump",
                31,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "FlowDetector": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/FlowDetector",
                164,
                None,
                ["inFlo", "NotInstalled", "PressureSwitch"],
                None,
                None,
                "ALL",
            ),
            "ProbeLocation": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/ProbeLocation",
                81,
                3,
                ["IntoPiping", "IntoTub"],
                None,
                2,
                "ALL",
            ),
            "CooldownTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/CooldownTime",
                35,
                "ALL",
            ),
            "MinSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/MinSetpointG",
                66,
                "ALL",
            ),
            "MaxSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/MaxSetpointG",
                68,
                "ALL",
            ),
            "EconType": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/EconType",
                70,
                None,
                ["Standard", "Night"],
                None,
                None,
                "ALL",
            ),
            "NoHeatPeriod": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/NoHeatPeriod",
                77,
                "ALL",
            ),
            "CoolZoneMode": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/CoolZoneMode",
                115,
                None,
                [
                    "CHILL",
                    "HEAT_W_BOOST",
                    "HEAT_SAVER",
                    "AUTO_W_BOOST",
                    "AUTO_SAVER",
                    "INTERNAL_HEAT",
                    "BOTH_HEAT",
                ],
                None,
                None,
                "ALL",
            ),
            "GenericHeatPumpID": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/GenericHeatPumpID",
                162,
                "ALL",
            ),
            "UL_CE": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/UL_CE",
                51,
                None,
                ["UL", "CE"],
                None,
                None,
                "ALL",
            ),
            "MaxNumberOfPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/MaxNumberOfPhases", 114, "ALL"
            ),
            "NbPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/NbPhases", 52, "ALL"
            ),
            "Max1phaseSinglePackCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Max1phaseSinglePackCurrent",
                165,
                "ALL",
            ),
            "Max2phaseSinglePackCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Max2phaseSinglePackCurrent",
                166,
                "ALL",
            ),
            "Max3phaseSinglePackCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Max3phaseSinglePackCurrent",
                167,
                "ALL",
            ),
            "Max1phaseDualPackCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Max1phaseDualPackCurrent",
                168,
                "ALL",
            ),
            "Max2phaseDualPackCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Max2phaseDualPackCurrent",
                169,
                "ALL",
            ),
            "Max3phaseDualPackCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Max3phaseDualPackCurrent",
                170,
                "ALL",
            ),
            "RealPackType": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/RealPackType",
                171,
                None,
                ["", "", "", "", "", "", "", "", "", "", "inYT", "", "inYJ"],
                None,
                None,
                None,
            ),
            "InputCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/InputCurrent", 53, "ALL"
            ),
            "MinimumInputCurrent": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/MinimumInputCurrent",
                122,
                "ALL",
            ),
            "InputMenu": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/InputMenu",
                74,
                None,
                ["Standard", "DualPack"],
                None,
                None,
                "ALL",
            ),
            "Out1Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out1Fuse",
                83,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out2Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out2Fuse",
                84,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out3Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out3Fuse",
                85,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out4Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out4Fuse",
                86,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out5Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out5Fuse",
                87,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out6Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out6Fuse",
                88,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out7Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out7Fuse",
                89,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out8Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out8Fuse",
                90,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out9Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out9Fuse",
                91,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out10Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out10Fuse",
                92,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out11Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out11Fuse",
                93,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Out12Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Out12Fuse",
                94,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Direct1Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Direct1Fuse",
                95,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "Direct2Fuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/Direct2Fuse",
                96,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "OutHtrFuse": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/OutHtrFuse",
                97,
                None,
                ["F1", "F2", "F3", "F21", "F22", "F23", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "F1Current": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/F1Current", 98, "ALL"
            ),
            "F2Current": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/F2Current", 99, "ALL"
            ),
            "F3Current": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/F3Current", 100, "ALL"
            ),
            "F21Current": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/F21Current", 101, "ALL"
            ),
            "F22Current": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/F22Current", 102, "ALL"
            ),
            "F23Current": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/F23Current", 103, "ALL"
            ),
            "F1Line": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/F1Line",
                104,
                None,
                ["", "", "", "", "", "", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "F2Line": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/F2Line",
                105,
                None,
                ["", "", "", "", "", "", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "F3Line": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/F3Line",
                106,
                None,
                ["", "", "", "", "", "", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "F21Line": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/F21Line",
                107,
                None,
                ["", "", "", "", "", "", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "F22Line": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/F22Line",
                108,
                None,
                ["", "", "", "", "", "", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "F23Line": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/F23Line",
                109,
                None,
                ["", "", "", "", "", "", "Line1", "Line2", "Line3"],
                None,
                None,
                "ALL",
            ),
            "FiltFreq": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltFreq", 3, "ALL"
            ),
            "FiltStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltStart",
                4,
                "ALL",
            ),
            "FiltDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltDur", 6, "ALL"
            ),
            "FiltDur2": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltDur2",
                116,
                "ALL",
            ),
            "EconStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStart",
                8,
                "ALL",
            ),
            "EconDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/EconDur", 10, "ALL"
            ),
            "EconProgAvailable": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconProgAvailable",
                71,
                None,
                ["NA", "STANDARD", "OUTSIDE_FILTER"],
                None,
                None,
                "ALL",
            ),
            "UDProgEcon": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/UDProgEcon",
                82,
                0,
                "ALL",
            ),
            "EconControlableManually": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconControlableManually",
                72,
                2,
                "ALL",
            ),
            "SoakOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SoakOnCustomKey",
                72,
                0,
                "ALL",
            ),
            "OffOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/OffOnCustomKey",
                72,
                1,
                "ALL",
            ),
            "CleanupOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/CleanupOnCustomKey",
                72,
                3,
                "ALL",
            ),
            "AuxOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/AuxOnCustomKey",
                72,
                4,
                "ALL",
            ),
            "QuickOnOffCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/QuickOnOffCustomKey",
                72,
                5,
                "ALL",
            ),
            "MultiKeyOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/MultiKeyOption",
                76,
                None,
                ["NoBlowerOnI2C", "BlowerOnI2C"],
                None,
                None,
                "ALL",
            ),
            "MasterSlave": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/MasterSlave",
                73,
                None,
                ["", "Master", "Slave"],
                None,
                None,
                "ALL",
            ),
            "SlaveConfig": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SlaveConfig",
                75,
                "ALL",
            ),
            "ProgOutAccessory": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/ProgOutAccessory",
                160,
                None,
                [
                    "NA",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "L120",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "ONZEN",
                    "",
                    "",
                    "",
                    "AUX",
                ],
                None,
                None,
                "ALL",
            ),
            "ProgOutStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/ProgOutStart",
                128,
                "ALL",
            ),
            "ProgOutDur": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/ProgOutDur",
                130,
                "ALL",
            ),
            "ProgOutFreq": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/ProgOutFreq",
                132,
                "ALL",
            ),
            "SilentMode": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SilentMode",
                157,
                None,
                ["NA", "OFF", "ECONOMY", "SLEEP", "NIGHT"],
                None,
                None,
                "ALL",
            ),
            "SilentDuration": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SilentDuration",
                158,
                "ALL",
            ),
            "SuspendSilentToRegulate": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SuspendSilentToRegulate",
                119,
                3,
                ["NOT_ALLOWED", "ALLOWED"],
                None,
                2,
                "ALL",
            ),
            "TimeFormat": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/TimeFormat",
                34,
                None,
                ["NA", "AmPm", "24h"],
                None,
                None,
                "ALL",
            ),
            "AmbiantOHTrigADC": GeckoWordStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/AmbiantOHTrigADC", 64, "ALL"
            ),
            "Pump1UserAccess": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/Pump1UserAccess",
                81,
                1,
                ["BothSpeeds", "HighSpeedOnly"],
                None,
                2,
                "ALL",
            ),
            "SelfCleanMsg": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/SelfCleanMsg", 81, 4, "ALL"
            ),
            "BlowerKeyOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/BlowerKeyOption",
                81,
                5,
                ["FreePumpKey", "LastPumpKey"],
                None,
                2,
                "ALL",
            ),
            "HeaterSoftStart": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/HeaterSoftStart", 81, 6, "ALL"
            ),
            "HeaterSoftStop": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/HeaterSoftStop", 81, 7, "ALL"
            ),
            "KeypadTherapySupport": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/KeypadTherapySupport",
                110,
                0,
                "ALL",
            ),
            "CustomKeyEnabled": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/CustomKeyEnabled", 110, 1, "ALL"
            ),
            "ConfigChange": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/ConfigChange",
                110,
                2,
                ["NO_RESTRICTION", "PASSWORD_PROTECTED"],
                None,
                2,
                "ALL",
            ),
            "BreakerChange": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/BreakerChange",
                110,
                3,
                ["NO_RESTRICTION", "PASSWORD_PROTECTED"],
                None,
                2,
                "ALL",
            ),
            "KeypadBacklightColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/KeypadBacklightColor",
                110,
                4,
                ["OFF", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE"],
                None,
                8,
                "ALL",
            ),
            "KeypadBacklightEdit": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/KeypadBacklightEdit",
                110,
                7,
                ["Disable", "Enable"],
                None,
                2,
                "ALL",
            ),
            "CustomerID": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/CustomerID",
                111,
                None,
                [
                    "Generic",
                    "Hydropool",
                    "EndlessPools",
                    "Wellis",
                    "Alps",
                    "Artesian",
                    "Arctic",
                    "Barefoot",
                    "Beachcomber",
                    "Bellagio",
                    "Leisure_Prod_Ind",
                    "Bullfrog",
                    "Coast",
                    "Dimension_one",
                    "Dynasty",
                    "Four_Wind",
                    "Hotspring",
                    "Jacuzzi",
                    "Jazzi",
                    "LA",
                    "Pro_Float",
                    "MAAX",
                    "Marquis",
                    "PDC",
                    "Premium_Leisure",
                    "Strong",
                    "Sunrans",
                    "Sunrise",
                    "SuperiorSpas",
                    "Spa_Industries",
                    "Viking",
                    "WWO_Whirlcare",
                    "Okeanos",
                    "Clearwater",
                    "Deluxe",
                    "Idol",
                    "Aspen",
                    "ThermoSpas",
                    "Titan_Spas",
                    "IberSpa",
                    "Master_Spas",
                ],
                None,
                None,
                "ALL",
            ),
            "ModeKeyAsInvertDisplayKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/ModeKeyAsInvertDisplayKey",
                112,
                1,
                "ALL",
            ),
            "InfoMsgConfig": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/InfoMsgConfig",
                112,
                2,
                ["HIDE_DETAILED_MSG", "SHOW_ALL_MSG", "", ""],
                None,
                4,
                "ALL",
            ),
            "LowerSetpointMenu": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/LowerSetpointMenu", 112, 4, "ALL"
            ),
            "SinglePumpKey": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/SinglePumpKey",
                112,
                6,
                ["NA", "ForSpeeds", "ForZones", ""],
                None,
                4,
                "ALL",
            ),
            "KeypadOptions4": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/KeypadOptions4", 113, "ALL"
            ),
            "DealerLockSupport": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/DealerLockSupport", 119, 0, "ALL"
            ),
            "LockEnabled": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/LockEnabled", 121, "ALL"
            ),
            "AssignVSP1": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/VSP/AssignVSP1",
                173,
                None,
                ["NA", "P1", "P2", "P3", "P4", "P5"],
                None,
                None,
                "ALL",
            ),
            "AssignVSP2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/VSP/AssignVSP2",
                174,
                None,
                ["NA", "P1", "P2", "P3", "P4", "P5"],
                None,
                None,
                "ALL",
            ),
            "AssignVSP3": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/VSP/AssignVSP3",
                175,
                None,
                ["NA", "P1", "P2", "P3", "P4", "P5"],
                None,
                None,
                "ALL",
            ),
            "AssignVSP4": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/VSP/AssignVSP4",
                176,
                None,
                ["NA", "P1", "P2", "P3", "P4", "P5"],
                None,
                None,
                "ALL",
            ),
            "AssignVSP5": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/VSP/AssignVSP5",
                177,
                None,
                ["NA", "P1", "P2", "P3", "P4", "P5"],
                None,
                None,
                "ALL",
            ),
            "ExerciseType": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExerciseType",
                133,
                None,
                ["NO_EXERCISE", "SWIM_EXERCISE"],
                None,
                None,
                "ALL",
            ),
            "ExerciseControl": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExerciseControl",
                134,
                None,
                ["USE_STANDARD_PUMPS", "USE_VARIABLE_SPEED_PUMPS"],
                None,
                None,
                "ALL",
            ),
            "ExercisePumpsUsed": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExercisePumpsUsed",
                135,
                0,
                [
                    "NO_PUMP",
                    "ALL_PUMPS",
                    "P2_TO_P5",
                    "P3_TO_P5",
                    "P4_AND_P5",
                    "P5_ONLY",
                    "",
                    "",
                ],
                None,
                8,
                "ALL",
            ),
            "ExerciseDedicatedPumps": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExerciseDedicatedPumps",
                135,
                7,
                "ALL",
            ),
            "ExerciseBuoyancyPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExerciseBuoyancyPump",
                178,
                None,
                ["NA", "P1", "P2", "P3", "P4", "P5"],
                None,
                None,
                "ALL",
            ),
            "ExerciseNumberOfIntensity": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExerciseNumberOfIntensity",
                136,
                "ALL",
            ),
            "ExcercisePumpIntensity1": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity1",
                137,
                "ALL",
            ),
            "ExcercisePumpIntensity2": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity2",
                139,
                "ALL",
            ),
            "ExcercisePumpIntensity3": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity3",
                141,
                "ALL",
            ),
            "ExcercisePumpIntensity4": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity4",
                143,
                "ALL",
            ),
            "ExcercisePumpIntensity5": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity5",
                145,
                "ALL",
            ),
            "ExcercisePumpIntensity6": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity6",
                147,
                "ALL",
            ),
            "ExcercisePumpIntensity7": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity7",
                149,
                "ALL",
            ),
            "ExcercisePumpIntensity8": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity8",
                151,
                "ALL",
            ),
            "ExcercisePumpIntensity9": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity9",
                153,
                "ALL",
            ),
            "ExcercisePumpIntensity10": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/Exercise/ExcercisePumpIntensity10",
                155,
                "ALL",
            ),
            "Zone1Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone1Led",
                123,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "Zone1Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone1Type",
                123,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "Zone2Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone2Led",
                124,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "Zone2Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone2Type",
                124,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "Zone3Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone3Led",
                125,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "Zone3Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone3Type",
                125,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "Zone4Led": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone4Led",
                126,
                0,
                ["RGB", "WHITE"],
                None,
                2,
                None,
            ),
            "Zone4Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/Zone4Type",
                126,
                1,
                ["NORMAL", "STATUS"],
                None,
                2,
                None,
            ),
            "NumberOfZones": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/NumberOfZones",
                127,
                0,
                ["", "1", "2", "3", "4"],
                None,
                8,
                None,
            ),
            "StatusLightNormalColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/StatusLightNormalColor",
                127,
                3,
                ["BLUE", "WHITE", "GREEN", "PURPLE", "YELLOW", "CYAN", "ORANGE", "OFF"],
                None,
                8,
                "ALL",
            ),
            "MappingEnable": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MappingEnable", 127, 7, None
            ),
            "MapZone1ToOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone1ToOut1", 123, 4, None
            ),
            "MapZone1ToOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone1ToOut2", 123, 5, None
            ),
            "MapZone1ToOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone1ToOut3", 123, 6, None
            ),
            "MapZone1ToOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone1ToOut4", 123, 7, None
            ),
            "MapZone2ToOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone2ToOut1", 124, 4, None
            ),
            "MapZone2ToOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone2ToOut2", 124, 5, None
            ),
            "MapZone2ToOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone2ToOut3", 124, 6, None
            ),
            "MapZone2ToOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone2ToOut4", 124, 7, None
            ),
            "MapZone3ToOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone3ToOut1", 125, 4, None
            ),
            "MapZone3ToOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone3ToOut2", 125, 5, None
            ),
            "MapZone3ToOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone3ToOut3", 125, 6, None
            ),
            "MapZone3ToOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone3ToOut4", 125, 7, None
            ),
            "MapZone4ToOut1": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone4ToOut1", 126, 4, None
            ),
            "MapZone4ToOut2": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone4ToOut2", 126, 5, None
            ),
            "MapZone4ToOut3": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone4ToOut3", 126, 6, None
            ),
            "MapZone4ToOut4": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/inMixConfig/MapZone4ToOut4", 126, 7, None
            ),
            "InMixZoneToFollow": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/InMixZoneToFollow",
                172,
                None,
                ["ZONE1", "ZONE2", "ZONE3", "ZONE4", "AS_12V_SUPPLY"],
                None,
                None,
                None,
            ),
            "StatusLightRemindersColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/StatusLightRemindersColor",
                179,
                4,
                [
                    "SAME_AS_NORMAL",
                    "WHITE",
                    "GREEN",
                    "PURPLE",
                    "YELLOW",
                    "CYAN",
                    "ORANGE",
                    "OFF",
                ],
                None,
                8,
                "ALL",
            ),
            "RemindersNotBlinkingLED": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/inMixConfig/RemindersNotBlinkingLED",
                179,
                7,
                None,
            ),
        }



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'MAS-IBC-32K v1'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 1

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "LL_Backrest": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Backrest",
                0,
                None,
                ["SINGLE", "DUAL", "DUAL_THERMOPLACE", "DUAL_LITESTREME"],
                None,
                None,
                None,
            ),
            "LL_Heater_1": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Heater_1",
                1,
                None,
                ["Intensity1", "Intensity2", "Intensity3", "Intensity4"],
                None,
                None,
                None,
            ),
            "LL_Heater_2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Heater_2",
                2,
                None,
                ["Intensity1", "Intensity2", "Intensity3", "Intensity4"],
                None,
                None,
                None,
            ),
            "LL_Blower": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Blower",
                3,
                None,
                ["STANDARD", "HIGH", "LOW"],
                None,
                None,
                None,
            ),
            "LL_Chromo": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Chromo",
                4,
                None,
                ["NONE", "BASIC", "DELUXE"],
                None,
                None,
                None,
            ),
            "LL_Audio": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Audio",
                5,
                None,
                ["NONE", "MP3"],
                None,
                None,
                None,
            ),
            "LL_Menu": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Menu",
                6,
                None,
                ["STANDARD", "HOTEL"],
                None,
                None,
                None,
            ),
            "LL_ComfortJet": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_ComfortJet",
                7,
                None,
                ["DEACTIVATE", "ACTIVATE"],
                None,
                None,
                None,
            ),
            "LL_Aromacloud": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/All/LL_Aromacloud",
                8,
                None,
                ["DEACTIVATE", "ACTIVATE"],
                None,
                None,
                None,
            ),
        }



"""A peekable async queue."""

import asyncio.queues
import collections
from typing import Any


class AsyncPeekableQueue:
    """A peekable async queue which allows queue filtering in consumers."""

    def __init__(self) -> None:
        """Initialize the async peekable queue."""
        self._marked = False
        self._normal_queue = collections.deque()
        self._command_queue = collections.deque()
        self._data_available = asyncio.Event()

    def _set_state(self) -> None:
        if self._command_queue or self._normal_queue:
            self._data_available.set()
        else:
            self._data_available.clear()

    def peek(self) -> Any:
        """Get the queue head or None."""
        if self._command_queue:
            return self._command_queue[0]
        if self._normal_queue:
            return self._normal_queue[0]
        return None

    @property
    def is_data_available(self) -> bool:
        """Is there any data available."""
        return self._data_available.is_set()

    @property
    def is_marked(self) -> bool:
        """Get the is_marked property."""
        return self._marked

    def pop(self) -> Any:
        """Pop the item off the top of the queue."""
        try:
            if self._command_queue:
                return self._command_queue.popleft()
            return self._normal_queue.popleft()
        finally:
            self._set_state()
            self._marked = False

    def push(self, item: Any) -> None:
        """Push an item on the bottom of the queue."""
        self._normal_queue.append(item)
        self._set_state()
        self._marked = False

    def push_command(self, item: Any) -> None:
        """Push an item on the bottom of the command queue."""
        self._command_queue.append(item)
        self._set_state()
        self._marked = False

    def mark(self) -> None:
        """Mark the queue."""
        self._marked = True

    async def wait(self) -> None:
        """Wait for there to be data queued."""
        await self._data_available.wait()

    def __bool__(self) -> bool:
        """Determine if there are any items in the queue."""
        return len(self) > 0

    def __len__(self) -> int:
        """Implement the __len__ function."""
        return self._command_queue.__len__() + self._normal_queue.__len__()



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoRemindersProtocolHandler,
    GeckoReminderType,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoRemindersHandler(unittest.TestCase):
    """Reminders handler tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoRemindersProtocolHandler.request(1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>REQRM\x01</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoRemindersProtocolHandler.req_response(
            [
                (GeckoReminderType.RINSE_FILTER, -13),
                (GeckoReminderType.CLEAN_FILTER, 257),
                (GeckoReminderType.CHANGE_WATER, 2),
                (GeckoReminderType.CHECK_SPA, 512),
                (GeckoReminderType.CHANGE_OZONATOR, 0),
                (GeckoReminderType.CHANGE_VISION_CARTRIDGE, 128),
                (GeckoReminderType.INVALID, 1),
            ],
            parms=PARMS,
        )
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>RMREQ"
            b"\x01\xf3\xff\x01"
            b"\x02\x01\x01\x01"
            b"\x03\x02\x00\x01"
            b"\x04\x00\x02\x01"
            b"\x05\x00\x00\x01"
            b"\x06\x80\x00\x01"
            b"\x00\x01\x00\x01"
            b"</DATAS></PACKT>",
        )

    def test_send_construct_set(self) -> None:
        handler = GeckoRemindersProtocolHandler.set(
            1,
            [
                (GeckoReminderType.RINSE_FILTER, -13),
                (GeckoReminderType.CLEAN_FILTER, 257),
                (GeckoReminderType.CHANGE_WATER, 2),
                (GeckoReminderType.CHECK_SPA, 512),
                (GeckoReminderType.CHANGE_OZONATOR, 0),
                (GeckoReminderType.CHANGE_VISION_CARTRIDGE, 128),
                (GeckoReminderType.INVALID, 1),
            ],
            parms=PARMS,
        )
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>SETRM\x01"
            b"\x01\xf3\xff\x01"
            b"\x02\x01\x01\x01"
            b"\x03\x02\x00\x01"
            b"\x04\x00\x02\x01"
            b"\x05\x00\x00\x01"
            b"\x06\x80\x00\x01"
            b"\x00\x01\x00\x01"
            b"</DATAS></PACKT>",
        )

    def test_send_construct_set_ack(self) -> None:
        handler = GeckoRemindersProtocolHandler.set_ack(parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>RMSET</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        self.assertTrue(handler.can_handle(b"REQRM", PARMS))
        self.assertTrue(handler.can_handle(b"RMREQ", PARMS))
        self.assertTrue(handler.can_handle(b"SETRM", PARMS))
        self.assertTrue(handler.can_handle(b"RMSET", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(b"REQRM\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)

    def test_recv_handle_response(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(
            b"RMREQ"
            b"\x01\x01\x00\x01\x02\x1f\x00\x01\x03\x29"
            b"\x00\x01\x04\xa9\x02\x01\x00\x01\x00\x00\x00\x01\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            PARMS,
        )
        self.assertListEqual(
            handler.reminders,
            [
                (GeckoReminderType.RINSE_FILTER, 1),
                (GeckoReminderType.CLEAN_FILTER, 31),
                (GeckoReminderType.CHANGE_WATER, 41),
                (GeckoReminderType.CHECK_SPA, 681),
                (GeckoReminderType.INVALID, 1),
                (GeckoReminderType.INVALID, 1),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
            ],
        )

    def test_recv_handle_boundary_conditions_response(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(
            b"RMREQ"
            b"\x01\xf3\xff\x01"
            b"\x02\x01\x01\x01"
            b"\x03\x02\x00\x01"
            b"\x04\x00\x02\x01"
            b"\x05\x00\x00\x01"
            b"\x06\x80\x00\x01"
            b"\x07\x01\x01\x01"
            b"\x08\x01\x01\x01"
            b"\xff\x01\x01\x01"
            b"\x00\x01\x00\x01",
            PARMS,
        )
        self.assertListEqual(
            handler.reminders,
            [
                (GeckoReminderType.RINSE_FILTER, -13),
                (GeckoReminderType.CLEAN_FILTER, 257),
                (GeckoReminderType.CHANGE_WATER, 2),
                (GeckoReminderType.CHECK_SPA, 512),
                (GeckoReminderType.CHANGE_OZONATOR, 0),
                (GeckoReminderType.CHANGE_VISION_CARTRIDGE, 128),
                (GeckoReminderType.INVALID, 1),
            ],
        )

    def test_recv_handle_negative_number_response(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(
            b"RMREQ\x01\xf3\xff\x01\x02\x11\x00\x01\x03/\x00\x01\x04\xaf\x02\x01"
            b"\x00\xf3\xff\x00\x00\xf3\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00",
            PARMS,
        )
        self.assertListEqual(
            handler.reminders,
            [
                (GeckoReminderType.RINSE_FILTER, -13),
                (GeckoReminderType.CLEAN_FILTER, 17),
                (GeckoReminderType.CHANGE_WATER, 47),
                (GeckoReminderType.CHECK_SPA, 687),
                (GeckoReminderType.INVALID, -13),
                (GeckoReminderType.INVALID, -13),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
            ],
        )

    def test_recv_handle_set_reminder(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(
            b"SETRM\x16\x01\x1e\x00\x01\x02\x00\x00\x01\x03/\x00\x01\x04\xaf\x02\x01\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00",
            PARMS,
        )
        self.assertListEqual(
            handler.reminders,
            [
                (GeckoReminderType.RINSE_FILTER, 30),
                (GeckoReminderType.CLEAN_FILTER, 0),
                (GeckoReminderType.CHANGE_WATER, 47),
                (GeckoReminderType.CHECK_SPA, 687),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
            ],
        )

    def test_recv_handle_set_reminder_reset_all(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(
            b"SETRMA\x01\x1e\x00\x01\x02<\x00\x01\x03Z\x00\x01\x04\xda\x02\x01\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00\x00\x1e\x00\x00",
            PARMS,
        )
        self.assertListEqual(
            handler.reminders,
            [
                (GeckoReminderType.RINSE_FILTER, 30),
                (GeckoReminderType.CLEAN_FILTER, 60),
                (GeckoReminderType.CHANGE_WATER, 90),
                (GeckoReminderType.CHECK_SPA, 730),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
                (GeckoReminderType.INVALID, 30),
            ],
        )

    def test_recv_handle_set_reminder_reset_sim(self) -> None:
        handler = GeckoRemindersProtocolHandler()
        handler.handle(
            b"SETRM\t\x01\xf3\xff\x01\x02\x00\x00\x01\x03\x00\x00\x01\x04\xaf\x02\x01\x00\xf3\xff\x01\x00\xf3\xff\x01\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01",
            PARMS,
        )
        self.assertListEqual(
            handler.reminders,
            [
                (GeckoReminderType.RINSE_FILTER, -13),
                (GeckoReminderType.CLEAN_FILTER, 0),
                (GeckoReminderType.CHANGE_WATER, 0),
                (GeckoReminderType.CHECK_SPA, 687),
                (GeckoReminderType.INVALID, -13),
                (GeckoReminderType.INVALID, -13),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
                (GeckoReminderType.INVALID, 0),
            ],
        )


if __name__ == "__main__":
    unittest.main()



"""
Complete sample client CUI - Console User Interface.

All the code to drive the CUI is in this file, it should only
talk to the facade as it is the example of how to integrate
geckolib into an automation system.
"""

import _curses
import asyncio
import curses
import curses.textpad
import logging
from datetime import UTC, datetime
from typing import Any, Self

from geckolib import (
    GeckoAsyncSpaDescriptor,
    GeckoAsyncSpaMan,
    GeckoConstants,
    GeckoSpaEvent,
)
from geckolib.automation.light import GeckoLight
from geckolib.automation.pump import GeckoPump
from geckolib.config import config_sleep
from geckolib.spa_state import GeckoSpaState

from .abstract_display import AbstractDisplay
from .config import Config

# Replace with your own UUID, see https://www.uuidgenerator.net/>
CLIENT_ID = "1eca3a27-9b00-476a-9645-d13f4b1f9b56"


_LOGGER = logging.getLogger(__name__)


class CUI(AbstractDisplay, GeckoAsyncSpaMan):
    """The console UI implementation."""

    @staticmethod
    def launch() -> None:
        """Launch the CUI."""
        curses.wrapper(CUI.main)

    @staticmethod
    def main(stdscr: curses.window) -> None:
        """Run the async loop."""
        curses.init_pair(1, curses.COLOR_RED, curses.COLOR_BLACK)
        asyncio.run(
            CUI.async_main(stdscr),
        )

    @staticmethod
    async def async_main(stdscr: curses.window) -> None:
        """Async main manages the console UI."""
        task = asyncio.current_task()
        if task is not None:
            task.set_name("CUI main")
            async with CUI(stdscr):
                pass

    def __init__(self, stdscr: _curses.window) -> None:
        """Initialize the CUI class."""
        AbstractDisplay.__init__(self, stdscr)
        GeckoAsyncSpaMan.__init__(self, CLIENT_ID)

        self._config = Config()
        self._spas: GeckoAsyncSpaDescriptor | None = None

        # Various flags based on the SpaMan events to simulate an
        # automation client
        self._can_use_facade = False
        self._lines: list[str | tuple[str, int]] = []

    async def __aenter__(self) -> Self:
        """Async enter for with statement."""
        await GeckoAsyncSpaMan.__aenter__(self)
        self.add_task(self._timer_loop(), "Timer", "CUI")
        await self.async_set_spa_info(
            self._config.spa_address, self._config.spa_id, self._config.spa_name
        )
        await self.run(self)
        return self

    async def __aexit__(self, *exc_info: object) -> None:
        """Async exit."""
        return await GeckoAsyncSpaMan.__aexit__(self, *exc_info)

    async def _timer_loop(self) -> None:
        try:
            while True:
                self.refresh()
                await config_sleep(1, "CUI Timer")
        except asyncio.CancelledError:
            _LOGGER.debug("Timer loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("Timer loop caught exception")
            raise

    async def handle_event(self, event: GeckoSpaEvent, **_kwargs: Any) -> None:
        """Rebuild the UI when there is an event."""
        try:
            _LOGGER.debug(f"{event} : {self.spa_state}")  # noqa: G004
            if event == GeckoSpaEvent.CLIENT_FACADE_IS_READY:
                self._can_use_facade = True
            elif event in (
                GeckoSpaEvent.CLIENT_FACADE_TEARDOWN,
                GeckoSpaState.ERROR_NEEDS_ATTENTION,
            ):
                self._can_use_facade = False

            self.refresh()
        except Exception:
            _LOGGER.exception("Exception in handle event")

    async def _select_spa(self, spa: GeckoAsyncSpaDescriptor) -> None:
        self._config.set_spa_id(spa.identifier_as_string)
        self._config.set_spa_address(spa.ipaddress)
        self._config.set_spa_name(spa.name)
        self._config.save()
        await self.async_set_spa_info(spa.ipaddress, spa.identifier_as_string, spa.name)

    async def _clear_spa(self) -> None:
        """Clear spa data."""
        self._config.set_spa_id(None)
        self._config.set_spa_address(None)
        self._config.set_spa_name(None)
        self._config.save()
        await self.async_set_spa_info(None, None, None)

    async def _select_next_watercare_mode(self) -> None:
        assert self.facade is not None  # noqa: S101
        assert self.facade.water_care is not None  # noqa: S101
        new_mode = (self.facade.water_care.active_mode + 1) % len(
            GeckoConstants.WATERCARE_MODE
        )
        await self.facade.water_care.async_set_mode(new_mode)

    async def increase_temp(self) -> None:
        """Increase the setpoint."""
        await self.facade.water_heater.async_set_target_temperature(
            self.facade.water_heater.target_temperature + 0.5
        )

    async def decrease_temp(self) -> None:
        """Decrease the setpoint."""
        await self.facade.water_heater.async_set_target_temperature(
            self.facade.water_heater.target_temperature - 0.5
        )

    async def key_press(self, keypad: int) -> None:
        """Simulate keypress."""
        await self.facade.spa.async_press(keypad)
        self.make_display()

    async def _device_click(self, device: GeckoPump | GeckoLight) -> None:
        _LOGGER.debug("Device %s clicked", device)
        try:
            if device.is_on:
                await device.async_turn_off()
            else:
                await device.async_turn_on()
        except Exception:
            _LOGGER.exception("Exception during click")

    def _located_spas(self, maxy: int, maxx: int) -> None:
        if self.spa_descriptors is None:
            return

        no_of_spas = len(self.spa_descriptors)
        if no_of_spas == 0:
            self._lines.append(
                (
                    "No spas were found on your network",
                    curses.color_pair(1),
                )
            )
            return

        buttons_height = no_of_spas * 3

        for idx, spa in enumerate(self.spa_descriptors):
            button_text = f"{idx + 1}. {spa.name} at {spa.ipaddress}"
            self.add_button(
                int((maxy - buttons_height) / 2) + (idx * 3),
                int(maxx / 2) - int(len(button_text) / 2) - 2,
                button_text,
                f"{idx + 1}",
                (self._select_spa, spa),
            )

    def _spa_problem(self) -> None:
        self._lines.append((f"{self.spa_name} not ready", curses.color_pair(1)))
        self._lines.append("")
        self._lines.append(f"{self.ping_sensor}")
        self._lines.append(f"{self.radio_sensor}")
        self._lines.append(f"{self.channel_sensor}")
        self._lines.append("")

        if self.spa_state == GeckoSpaState.ERROR_RF_FAULT:
            self._lines.append(
                (
                    "Lost contact with your spa, it looks as if it is turned off",
                    curses.color_pair(1),
                )
            )
        if self.spa_state == GeckoSpaState.ERROR_PING_MISSED:
            self._lines.append(
                (
                    "Lost contact with your intouch2 module, please investigate",
                    curses.color_pair(1),
                )
            )

    def _build_facade_ui(self, _maxy: int, maxx: int) -> None:  # noqa: PLR0912, PLR0915
        if self.facade is None:
            return
        curx = 2
        cury = 1
        button_line = []

        for device in list(
            self.facade.pumps
            + self.facade.blowers
            + self.facade.lights
            + self.facade.switches
        ):
            device_button = self.add_button(
                cury,
                curx,
                [device.name, device.state],
                None,
                (self._device_click, device),
            )
            if device_button.x + device_button.width > maxx - 2:
                delta = int(((maxx - 2) - curx) / 2)
                for moveit in button_line:
                    moveit.move(moveit.y, moveit.x + delta)
                button_line = []
                curx = 2
                cury += 4
                device_button.move(cury, curx)
            curx += device_button.width + 1
            button_line.append(device_button)

        delta = int(((maxx - 2) - curx) / 2)
        for moveit in button_line:
            moveit.move(moveit.y, moveit.x + delta)

        if self.facade.mrsteam.is_available:
            self._lines.append("We have a Mr Steam unit")
            self._lines.append("")
        if self.facade.bainultra.is_available:
            self._lines.append("We have a Bain Ultra bath")
            self._lines.append("")
            self._lines.append(f"{self.facade.bainultra.bath_runtime}")
            self._lines.append(f"{self.facade.bainultra.bath_intensity}")
            self._lines.append(f"{self.facade.bainultra.backrest}")
            self._lines.append(f"{self.facade.bainultra.backrest2}")
            if self.facade.bainultra.chroma.is_available:
                self._lines.append(f"{self.facade.bainultra.chroma}")
            self._lines.append(f"{self.facade.bainultra.drying_cycle}")
            self._lines.append(f"{self.facade.bainultra.drying_cycle_delay}")

        if self.facade.water_heater.is_available:
            self._lines.append(f"{self.facade.water_heater}")
        if self.facade.inmix.is_available:
            self._lines.extend(f"{zone}" for zone in self.facade.inmix.zones)
            if self.facade.inmix.syncro.is_available:
                self._lines.append(f"{self.facade.inmix.syncro}")
        if self.facade.reminders_manager.is_available:
            self._lines.extend(
                f"{reminder}" for reminder in self.facade.reminders_manager.reminders
            )
        # if self.facade.water_care.is_available:
        if self.facade.water_care.is_available:
            self._lines.append(f"{self.facade.water_care}")
        if self.facade.heatpump.is_available:
            self._lines.append(f"{self.facade.heatpump}")
        if self.facade.ingrid.is_available:
            self._lines.append(f"{self.facade.ingrid}")
        if self.facade.lockmode.is_available:
            self._lines.append(f"{self.facade.lockmode}")
        self._lines.extend(
            f"{sensor}"
            for sensor in [
                *self.facade.sensors,
                *self.facade.binary_sensors,
            ]
        )
        if self.facade.eco_mode is not None:
            self._lines.append(f"{self.facade.eco_mode}")
        self._lines.append(
            f"{self.ping_sensor} (Responding: {self._spa.is_responding_to_pings})"
        )
        self._lines.append(f"{self.radio_sensor}")
        self._lines.append(f"{self.channel_sensor}")
        self._lines.append(f"{self.facade.error_sensor}")
        self._lines.append(f"{self.facade.spa_in_use_sensor}")

    def make_display(self) -> None:
        """Make a display."""
        try:
            maxy, maxx = self.stdscr.getmaxyx()
            self.stdscr.box()

            title = (
                f"{self.facade.name} is ready"
                if self._can_use_facade
                else "Gecko Async Sample App"
            )
            self.stdscr.addstr(0, int((maxx - len(title)) / 2), title)

            self._lines = []

            if self._can_use_facade:
                self._build_facade_ui(maxy, maxx)

            elif self.spa_state == GeckoSpaState.LOCATED_SPAS:
                self._located_spas(maxy, maxx)

            elif self.spa_state in (
                GeckoSpaState.ERROR_RF_FAULT,
                GeckoSpaState.ERROR_PING_MISSED,
                GeckoSpaState.ERROR_NEEDS_ATTENTION,
            ):
                self._spa_problem()

            self._lines.append("")

            if self._can_use_facade:
                assert self.facade is not None  # noqa: S101
                if self.facade.water_heater.is_available:
                    self._lines.append("Press '+' to increase setpoint")
                    self._commands["+"] = self.increase_temp
                    self._lines.append("Press '-' to decrease setpoint")
                    self._commands["-"] = self.decrease_temp
                if self.facade.water_care.is_available:
                    self._lines.append("Press 'w' to select next watercare mode")
                    self._commands["w"] = self._select_next_watercare_mode

            #       ┌──────┐
            #       │ eXit │
            #       └──────┘
            _exit_button = self.add_button(
                maxy - 4, maxx - 10, "eXit", "x", self.set_exit
            )

            #       ┌─────────┐    ┌─────────────┐
            #       │ Re-scan │ or │ Re-connnect │
            #       └─────────┘    └─────────────┘
            reset_text = (
                "Re-scan"
                if self.spa_state
                in (
                    GeckoSpaState.LOCATING_SPAS,
                    GeckoSpaState.LOCATED_SPAS,
                )
                and self._config.spa_id is None
                else "Re-connect"
            )
            _reset_button = self.add_button(
                maxy - 4,
                _exit_button.x - len(reset_text) - 5,
                reset_text,
                "r",
                self.async_reset,
            )

            #       ┌───────────┐
            #       │ Clear Spa │
            #       └───────────┘
            if self._config.spa_id is not None:
                _clear_spa = self.add_button(
                    maxy - 4, _reset_button.x - 14, "Clear Spa", "c", self._clear_spa
                )

            # Lines of data centred

            half = int(len(self._lines) / 2)

            for idx, line in enumerate(self._lines):
                attr: int = curses.A_NORMAL
                text: str = line
                if isinstance(line, tuple):
                    attr = int(line[1])
                    text = line[0]
                self.stdscr.addstr(
                    int(maxy / 2) - half + idx, int((maxx - len(text)) / 2), text, attr
                )

            # Status liine
            self.stdscr.addstr(
                maxy - 2,
                1,
                f"{datetime.now(tz=UTC):%x %X} - {self}",
            )

        except _curses.error:
            # If window gets too small, we won't output anything
            _LOGGER.warning("Window too small")
            self.stdscr.erase()
            self.stdscr.addstr("Window too small")

        except Exception:
            _LOGGER.exception("During make display")



"""Gecko CURCH/CHCUR handlers."""

from __future__ import annotations

import logging
import struct
from typing import Any

from geckolib.config import GeckoConfig

from .packet import GeckoPacketProtocolHandler

CURCH_VERB = b"CURCH"
CHCUR_VERB = b"CHCUR"
GETCHANNEL_FORMAT = ">BB"

_LOGGER = logging.getLogger(__name__)


class GeckoGetChannelProtocolHandler(GeckoPacketProtocolHandler):
    """Handle CURCH/CHCUR verbs."""

    @staticmethod
    def request(seq: int, **kwargs: Any) -> GeckoGetChannelProtocolHandler:
        """Generate request."""
        return GeckoGetChannelProtocolHandler(
            content=b"".join([CURCH_VERB, struct.pack(">B", seq)]),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def response(
        channel: int, signal_strength: int, **kwargs: Any
    ) -> GeckoGetChannelProtocolHandler:
        """Generate response."""
        return GeckoGetChannelProtocolHandler(
            content=b"".join(
                [
                    CHCUR_VERB,
                    struct.pack(
                        GETCHANNEL_FORMAT,
                        channel,
                        signal_strength,
                    ),
                ]
            ),
            **kwargs,
        )

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        self.channel = self.signal_strength = None

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle this verb."""
        return received_bytes.startswith((CURCH_VERB, CHCUR_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        if received_bytes.startswith(CURCH_VERB):
            self._sequence = struct.unpack(">B", remainder)[0]
            return  # Stay in the handler list
        # Otherwise must be CHCUR
        (
            self.channel,
            self.signal_strength,
        ) = struct.unpack(GETCHANNEL_FORMAT, remainder)
        self._should_remove_handler = True



"""Unit tests for the structure accessor class."""  # noqa: INP001

from typing import Any

import pytest
from context import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoWordStructAccessor,
)

from geckolib.driver.accessor import GeckoStructAccessor


class MockStructure(GeckoAsyncStructure):
    """Wrapper to help testing."""

    last_pos: int | None = None
    last_len: int | None = None
    last_data: bytes | None = None
    last_value: Any | None = None

    def __init__(self) -> None:
        """Initialize the mock structure."""
        super().__init__(self.set_value_cb)
        self.set_status_block(
            # Bytes 0-16 are identity values
            b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10"
            # Bytes 17-18 are a temperature in farenheit 10ths from freezing
            b"\x02\xbe"
            # Bytes 19-20 are a sized bitpos enum
            b"\x11\x70"
            # Bytes 21-22 are an empty sized bitpos enum
            b"\x00\x00"
            # Byte 23 is a sparse select value (40)
            b"\x28"
        )

    async def set_value_cb(self, pos: int, len_: int, newvalue: Any) -> None:
        self.last_pos = pos
        self.last_len = len_
        self.last_data = GeckoStructAccessor.pack_data(len_, newvalue)
        self.replace_status_block_segment(pos, self.last_data)
        self.last_value = newvalue


class TestStructAccessor:
    """Test the GeckoStructAccessor class."""

    struct = MockStructure()

    def test_read_byte(self) -> None:
        """Can we read a byte from the structure."""
        accessor = GeckoByteStructAccessor(self.struct, "PackBootRev", 5, None)
        assert accessor.value == 5

    @pytest.mark.asyncio
    async def test_write_byte_fails(self) -> None:
        """Can we write a byte to the structure without the RW tag."""
        accessor = GeckoByteStructAccessor(self.struct, "PackBootRev", 5, None)
        with pytest.raises(AttributeError):
            await accessor.async_set_value(6)

    @pytest.mark.asyncio
    async def test_write_byte(self) -> None:
        """Can we write a byte to the structure."""
        accessor = GeckoByteStructAccessor(self.struct, "PackBootRev", 5, "ALL")
        await accessor.async_set_value(6)
        assert self.struct.last_pos == 5
        assert self.struct.last_data == b"\x06"

    @pytest.mark.asyncio
    async def test_write_byte_from_string(self) -> None:
        """Can we write a byte (as a string) to the structure."""
        accessor = GeckoByteStructAccessor(self.struct, "PackBootRev", 5, "ALL")
        await accessor.async_set_value("6")
        assert self.struct.last_pos == 5
        assert self.struct.last_data == b"\x06"

    def test_read_word(self) -> None:
        """Can we read a word from the structure."""
        accessor = GeckoWordStructAccessor(self.struct, "RealSetPointG", 17, None)
        assert accessor.value == 702

    @pytest.mark.asyncio
    async def test_write_word(self) -> None:
        """Can we write a word to the structure."""
        accessor = GeckoWordStructAccessor(self.struct, "RealSetPointG", 17, "All")
        await accessor.async_set_value(726)
        assert self.struct.last_pos == 17
        assert self.struct.last_data == b"\x02\xd6"

    @pytest.mark.asyncio
    async def test_write_word_as_string(self) -> None:
        """Can we write a word (as a string) to the structure."""
        accessor = GeckoWordStructAccessor(self.struct, "RealSetPointG", 17, "All")
        await accessor.async_set_value("726")
        assert self.struct.last_pos == 17
        assert self.struct.last_data == b"\x02\xd6"

    def test_read_enum(self) -> None:
        """Can we read an enum from the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct,
            "PackType",
            6,
            None,
            [
                "Unknown",
                "inXE",
                "MasIBC",
                "MIA",
                "DJS4",
                "inClear",
                "inXM",
                "K600",
                "inTerface",
                "inTouch",
                "inYT",
                "K800",
                "inYJ",
            ],
            None,
            None,
            None,
        )
        assert accessor.value == "inXM"

    @pytest.mark.asyncio
    async def test_write_enum(self) -> None:
        """Can we write an enum to the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct,
            "PackType",
            6,
            None,
            (
                "Unknown|inXE|MasIBC|MIA|DJS4|inClear|inXM|K600"
                "|inTerface|inTouch|inYT|K800|inYJ"
            ),
            None,
            None,
            "All",
        )
        await accessor.async_set_value("inYJ")
        assert self.struct.last_pos == 6
        assert self.struct.last_data == b"\x0c"

    @pytest.mark.asyncio
    async def test_write_enum_not_member(self) -> None:
        """Can we write an enum to the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct,
            "PackType",
            6,
            None,
            (
                "Unknown|inXE|MasIBC|MIA|DJS4|inClear|inXM|K600"
                "|inTerface|inTouch|inYT|K800|inYJ"
            ),
            None,
            None,
            "All",
        )
        with pytest.raises(ValueError, match="Not A Member"):
            await accessor.async_set_value("Not A Member")

    def test_read_bool(self) -> None:
        """Can we read a bool from the structure."""
        accessor = GeckoBoolStructAccessor(self.struct, "RelayStuck", 2, 6, None)
        assert not accessor.value

    @pytest.mark.asyncio
    async def test_write_bool(self) -> None:
        """Can we write a bool to the structure."""
        accessor = GeckoBoolStructAccessor(self.struct, "RelayStuck", 2, 6, "All")
        await accessor.async_set_value(True)  # noqa: FBT003
        assert self.struct.last_pos == 2
        assert self.struct.last_data == b"B"

    @pytest.mark.asyncio
    async def test_write_bool_as_string(self) -> None:
        """Can we write a bool (as a string) to the structure."""
        accessor = GeckoBoolStructAccessor(self.struct, "RelayStuck", 2, 6, "All")
        await accessor.async_set_value("True")
        assert self.struct.last_pos == 2
        assert self.struct.last_data == b"B"

    def test_read_bitpos_enum(self) -> None:
        """Can we read a bitpos enum from the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct, "UdP3", 3, 4, "OFF|LO|HI", None, 4, None
        )
        assert accessor.value == "OFF"
        assert accessor.raw_value == 0

    @pytest.mark.asyncio
    async def test_write_bitpos_enum(self) -> None:
        """Can we write an enum to the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct, "UdP3", 3, 4, "OFF|LO|HI", None, 4, "All"
        )
        await accessor.async_set_value("HI")
        assert self.struct.last_pos == 3
        assert self.struct.last_data == b"#"
        assert accessor.value == "HI"
        assert accessor.raw_value == 2
        await accessor.async_set_value("LO")
        assert self.struct.last_data == b"\x13"
        assert accessor.raw_value == 1
        await accessor.async_set_value("OFF")
        assert self.struct.last_data == b"\x03"
        assert accessor.raw_value == 0

    def test_read_sized_bitpos_enum(self) -> None:
        """Can we read a sized bitpos enum from the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct, "UdP2", 19, 12, "OFF|LO|HI", 2, 4, None
        )
        assert accessor.value == "LO"

    @pytest.mark.asyncio
    async def test_write_sized_bitpos_enum(self) -> None:
        """Can we write a sized bitpos enum to the structure."""
        accessor = GeckoEnumStructAccessor(
            self.struct, "UdP2", 19, 12, "OFF|LO|HI", 2, 4, "All"
        )
        await accessor.async_set_value("HI")
        assert self.struct.last_pos == 19
        assert self.struct.last_data is not None
        assert chr(0b00100001) + chr(0b01110000) == self.struct.last_data.decode(
            "latin-1"
        )
        await accessor.async_set_value("LO")
        assert chr(0b00010001) + chr(0b01110000) == self.struct.last_data.decode(
            "latin-1"
        )
        await accessor.async_set_value("OFF")
        assert chr(0b00000001) + chr(0b01110000) == self.struct.last_data.decode(
            "latin-1"
        )

    @pytest.mark.asyncio
    async def test_multiple_write_bitpos_enum(self) -> None:
        """Can we write multiple bitpos enums to the structure."""
        accessor_p1 = GeckoEnumStructAccessor(
            self.struct, "UdP1", 21, 14, "OFF|LO|HI", 2, 4, "ALL"
        )
        accessor_p2 = GeckoEnumStructAccessor(
            self.struct, "UdP2", 21, 12, "OFF|LO|HI", 2, 4, "All"
        )

        await accessor_p2.async_set_value("HI")
        assert self.struct.last_pos == 21
        assert self.struct.last_data == b"\x20\x00"
        assert self.struct.last_len == 2
        assert self.struct.last_data is not None
        assert len(self.struct.last_data) == 2

        await accessor_p1.async_set_value("HI")
        assert self.struct.last_pos == 21
        assert self.struct.last_data == b"\xa0\x00"
        assert self.struct.last_len == 2
        assert len(self.struct.last_data) == 2

    def test_sparse_enum_read(self) -> None:
        dry_delay = GeckoEnumStructAccessor(
            self.struct,
            "LogStructure/UserDemands/UserDryingDelay",
            23,
            0,
            [
                "0",
                "",
                "",
                "",
                "",
                "5",
                "",
                "",
                "",
                "",
                "10",
                "",
                "",
                "",
                "",
                "15",
                "",
                "",
                "",
                "",
                "20",
                "",
                "",
                "",
                "",
                "25",
                "",
                "",
                "",
                "",
                "30",
                "",
                "",
                "",
                "",
                "35",
                "",
                "",
                "",
                "",
                "40",
                "",
                "",
                "",
                "",
                "45",
                "",
                "",
                "",
                "",
                "50",
                "",
                "",
                "",
                "",
                "55",
                "",
                "",
                "",
                "",
                "60",
            ],
            None,
            63,
            "ALL",
        )

        assert len(dry_delay.items) == 61
        assert dry_delay.raw_value == 40
        assert dry_delay.value == "40"
